//remove shader5!
/*************************************************************
*Shaders & demo by: Jason Gorski
*Email: jasejc@aol.com
*CS594 University of Illinios at Chicago
*
*LED Billboard Tutorial
*For more information about this shader view the tutorial page
*at http://www2.uic.edu/~jgorsk2 or email me
*************************************************************/ 

//basic includes
#include "stdafx.h"

//openGl related
#define GLEW_STATIC 1
#include <GL/glew.h>
#include <GL/glut.h>
#include <GL/glaux.h>

//Model loading
#include "obj.h"
int billboardObjFile;

//used in shader initialization
GLhandleARB v,f,billboard_shader,
	billboard1_shader,billboard2_shader,billboard3_shader,billboard4_shader,billboard5_shader;

//used to cycle through billboard shaders
#define BILLBOARD_TYPE_COUNT 4
enum { BILL_I, BILL_II, BILL_III, BILL_IV, BILL_V };
int current_type;

//used to store useful data associated with each billboard texture
struct BILLBOARD_TYPE {
	int lcf;//least common factor of x and y
	int x,y;//width = x, height = y
};

//used to dynamically change shader uniforms
int pixel_size = 0;
float tolerance = 0.3;
float pixel_radius = 0.6;
int luminance_steps = 12;
float luminance_boost = 0.0;
float color_boost = 0.03;
float burnt_out_percent = 0.2;

//Texture indexing variables and texture capping variables
//used for cycling through billboard textures
int current_texture;
#define TEXTURE_COUNT 14
#define MAX_BILLBOARD_TEXTURES 6
int billboard_texture_count = MAX_BILLBOARD_TEXTURES;
//billboard textures
#define FLAG_ID	0
#define WWBD_ID 1
#define HOMESTAR_ID 2
#define LOGO_ID 3
#define NOAH_ID 4
#define NEON_INDEX 5
//other textures
#define NOISE_ID 6
#define DAY_GRASS_ID 7
#define DAY_ID 8
#define NIGHT_GRASS_ID 9
#define NIGHT_ID 10
#define METAL_ID 11
#define NEON_ON_ID 12
#define NEON_OFF_ID 13
UINT textures[TEXTURE_COUNT] = {0};

//store a billboard struct for each texture
BILLBOARD_TYPE billboards[MAX_BILLBOARD_TEXTURES];

//View control and mouse tracking variables
#include "Camera.h"
float speed = 0.5;
CCamera myView;
int xPrev = 0, yPrev = 0;
bool dragging = false;

//bools for drawing some specific examples of using the final billboard texture
bool drawExample1, drawExample2;

using namespace std;

/**************************************Utility functions***********************************/
//read the shader file and return a pointer to it
char* readShaderFile(char* filename)
{
	ifstream file (filename, ios::in|ios::binary|ios::ate);

	if(file)
	{
		int count = file.tellg();
		char* shader = new char[count+1];
		
		file.seekg(0,ios::beg);
		file.read(shader,count);
		shader[count] = '\0';

		return shader;
	}
	return NULL;
}

//apply uniforms for the current billboard shader
void applyShader(int shader)
{
	GLint uniform;
	glUseProgramObjectARB(shader);
	switch(current_type) {					
		case BILL_V:

		/**********************added since billboard4*************************/
		SYSTEMTIME st;
		GetSystemTime(&st);
		uniform = glGetUniformLocationARB(shader, "time"); 
		glUniform1iARB(uniform, int(st.wMilliseconds));
		/*********************************************************************/

		case BILL_IV:

		/**********************added since billboard3*************************/
		uniform = glGetUniformLocationARB(shader, "colorBoost"); 
		glUniform1fARB(uniform, color_boost);
		uniform = glGetUniformLocationARB(shader, "burntOutPercent"); 
		glUniform1fARB(uniform, burnt_out_percent);
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glBindTexture(GL_TEXTURE_2D, textures[NOISE_ID]);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);
		uniform = glGetUniformLocationARB(shader, "noiseTexture"); 
		glUniform1iARB(uniform, 1);
		/*********************************************************************/

		case BILL_III:

		/**********************added since billboard2*************************/
		uniform = glGetUniformLocationARB(shader, "luminanceSteps"); 
		glUniform1iARB(uniform, luminance_steps);
		uniform = glGetUniformLocationARB(shader, "luminanceBoost"); 
		glUniform1fARB(uniform, luminance_boost);
		/*********************************************************************/

		case BILL_II:

		/**********************added since billboard1*************************/
		uniform = glGetUniformLocationARB(shader, "tolerance"); 
		glUniform1fARB(uniform, tolerance);
		uniform = glGetUniformLocationARB(shader, "pixelRadius"); 
		glUniform1fARB(uniform, pixel_radius);
		/*********************************************************************/

		case BILL_I:

		uniform = glGetUniformLocationARB(shader, "pixelSize"); 
		glUniform1iARB(uniform, pixel_size);
		uniform = glGetUniformLocationARB(shader, "billboardSize"); 
		glUniform2iARB(uniform, billboards[current_texture].x, billboards[current_texture].y);

		glActiveTextureARB(GL_TEXTURE0_ARB);
		glBindTexture(GL_TEXTURE_2D, textures[current_texture]);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		uniform = glGetUniformLocationARB(shader, "billboardTexture"); 
		glUniform1iARB(uniform, 0);
		break;

		default:

		glUseProgramObjectARB(0);//no shader
		glBindTexture(GL_TEXTURE_2D, 0);
		break;
	}
}

/*****************************Functions mapped into GLUT system****************************/

//this is our glut display method, it updates the screen
void display()
{
	glViewport(0,0, glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT));
    
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();

	myView.look();//Camera object to orient us to current view

	glPushMatrix();
	  if(drawExample1) { //1st of 2 presets
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f,glutGet(GLUT_WINDOW_WIDTH)/glutGet(GLUT_WINDOW_HEIGHT), 0.9f ,300.0f);
		
		glMatrixMode(GL_MODELVIEW);

		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,textures[DAY_GRASS_ID]);

		glPushMatrix();
			glScalef(2.0,2.0,2.0);
			glBegin(GL_QUADS);//draw grass
			{
				glTexCoord2f(0.0,0.0);
				glVertex3f(-60.0,0.0,-80.0);
				glTexCoord2f(0.0,1.0);
				glVertex3f(-60.0,0.0,80.0);
				glTexCoord2f(1.0,1.0);
				glVertex3f(60.0,0.0,80.0);
				glTexCoord2f(1.0,0.0);
				glVertex3f(60.0,0.0,-80.0);
			}
			glEnd();

			glBindTexture(GL_TEXTURE_2D,textures[DAY_ID]);
			glBegin(GL_QUADS);//draw background quad
			{
				glTexCoord2f(1.0,0.0);
				glVertex3f(60.0,-20.0,-50.0);
				glTexCoord2f(1.0,1.0);
				glVertex3f(60.0,60,-50.0);
				glTexCoord2f(0.0,1.0);
				glVertex3f(-60.0,60,-50.0);
				glTexCoord2f(0.0,0.0);
				glVertex3f(-60.0,-20,-50.0);
			}
			glEnd();
		glPopMatrix();

		glBindTexture(GL_TEXTURE_2D,textures[METAL_ID]);
		obj_draw_file(billboardObjFile);

		applyShader(billboard_shader);
		glBegin(GL_QUADS);//draw billboard quad
		{
			glTexCoord2f(1.0,0.0);
			glVertex3f(20.0,20.0-10.0,0.0);
			glTexCoord2f(1.0,1.0);
			glVertex3f(20.0,20.0+10.0,0.0);
			glTexCoord2f(0.0,1.0);
			glVertex3f(-20.0,20.0+10.0,0.0);
			glTexCoord2f(0.0,0.0);
			glVertex3f(-20.0,20.0-10.0,0.0);
		}
		glEnd();
		applyShader(0);
	  }
	  else if(drawExample2) { //2nd of 2 presets
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f,glutGet(GLUT_WINDOW_WIDTH)/glutGet(GLUT_WINDOW_HEIGHT), 0.9f ,300.0f);
		
		glMatrixMode(GL_MODELVIEW);

		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,textures[NIGHT_GRASS_ID]);
		glPushMatrix();
			glScalef(2.0,2.0,2.0);
			glBegin(GL_QUADS);//draw grass
			{
				glTexCoord2f(0.0,0.0);
				glVertex3f(-60.0,0.0,-80.0);
				glTexCoord2f(0.0,1.0);
				glVertex3f(-60.0,0.0,80.0);
				glTexCoord2f(1.0,1.0);
				glVertex3f(60.0,0.0,80.0);
				glTexCoord2f(1.0,0.0);
				glVertex3f(60.0,0.0,-80.0);
			}
			glEnd();

			glBindTexture(GL_TEXTURE_2D,textures[NIGHT_ID]);
			glBegin(GL_QUADS);//draw background quad
			{
				glTexCoord2f(1.0,0.0);
				glVertex3f(60.0,-20.0,-50.0);
				glTexCoord2f(1.0,1.0);
				glVertex3f(60.0,60,-50.0);
				glTexCoord2f(0.0,1.0);
				glVertex3f(-60.0,60,-50.0);
				glTexCoord2f(0.0,0.0);
				glVertex3f(-60.0,-20,-50.0);
			}
			glEnd();
		glPopMatrix();

		glBindTexture(GL_TEXTURE_2D,textures[METAL_ID]);
		obj_draw_file(billboardObjFile);

		applyShader(billboard_shader);
		glBegin(GL_QUADS);//draw billboard quad
		{
			glTexCoord2f(1.0,0.0);
			glVertex3f(20.0,20.0-10.0,0.0);
			glTexCoord2f(1.0,1.0);
			glVertex3f(20.0,20.0+10.0,0.0);
			glTexCoord2f(0.0,1.0);
			glVertex3f(-20.0,20.0+10.0,0.0);
			glTexCoord2f(0.0,0.0);
			glVertex3f(-20.0,20.0-10.0,0.0);
		}
		glEnd();
		applyShader(0);
	  }
	  else { //default case, allows for cycling through different textures & billboard shaders
		if(current_texture == NEON_INDEX) {
			SYSTEMTIME st;
			GetSystemTime(&st);
			if((st.wSecond % 2) == 0.0)
				textures[NEON_INDEX] = textures[NEON_ON_ID];
			else
				textures[NEON_INDEX] = textures[NEON_OFF_ID];
			glutPostRedisplay();
		}

		applyShader(billboard_shader);
		float xyRatio = billboards[current_texture].x/billboards[current_texture].y;
		glBegin(GL_QUADS);
		{
			glTexCoord2f(1.0,0.0);
			glVertex3f(20.0,15.0-20.0*(1/xyRatio),0.0);
			glTexCoord2f(1.0,1.0);
			glVertex3f(20.0,15.0+20.0*(1/xyRatio),0.0);
			glTexCoord2f(0.0,1.0);
			glVertex3f(-20.0,15.0+20.0*(1/xyRatio),0.0);
			glTexCoord2f(0.0,0.0);
			glVertex3f(-20.0,15.0-20.0*(1/xyRatio),0.0);
		}
	    glEnd();
		applyShader(0);
	  }
	 glPopMatrix();

	glFlush();
	glutSwapBuffers();
}

//standard reshapping function
void changeSize(int w, int h) {

	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if(h == 0)
		h = 1;

	float ratio = 1.0* w / h;

	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	// Set the viewport to be the entire window
    glViewport(0, 0, w, h);

	// Set the correct perspective.
	gluPerspective(45, ratio, .9, 100);
	glMatrixMode(GL_MODELVIEW);
}

//mapped to keyboard, camera moves FPS-style, lots of shader uniforms are adjustable through key presses
void keyHandler(unsigned char key, int x, int y)
{
	switch(key)
	{
	case 'l':
		printf("%f %f %f \n\tat %f %f %f \n\tup %f %f %f",
			myView.position().x,myView.position().y,myView.position().z,
			myView.view().x,myView.view().y,myView.view().z,
			myView.upVector().x,myView.upVector().y,myView.upVector().z);
		break;
	case 'w': //move forward
		myView.move(speed);
		break;
	case 's': //move back
		myView.move(-speed);
		break;
	case 'a': //strafe left
		myView.strafe(-speed);
		break;
	case 'd': //strafe right
		myView.strafe(speed);
		break;
	case '-': case '_': //decrement size of pixel regions
		pixel_size -= billboards[current_texture].lcf;
		if(pixel_size <= 0)
			pixel_size = billboards[current_texture].lcf;
		printf("pixelSize = %i\n",pixel_size);
		break;
	case '=': case '+': //increment size of pixel regions
		pixel_size += billboards[current_texture].lcf;
		if(pixel_size > min(billboards[current_texture].x,billboards[current_texture].y))
			pixel_size = min(billboards[current_texture].x,billboards[current_texture].y);
		printf("pixelSize = %i\n",pixel_size);
		break;
	case '[': case '{': //decrement size of pixel region radius
		pixel_radius -= 0.005;
		if(pixel_radius < 0.0)
			pixel_radius = 0.0;
		printf("pixelRadius [0.0,1.0] = %f\n",pixel_radius);
		break;
	case ']': case '}': //increment size of pixel region radius
		pixel_radius += 0.005;
		if(pixel_radius > 1.0)
			pixel_radius = 1.0;
		printf("pixelRadius [0.0,1.0] = %f\n",pixel_radius);
		break;
	case '9': case '(': //decrement size of tolerance on pixel region radius edges
		tolerance -= 0.005;
		if(tolerance < 0.0)
			tolerance = 0.0;
		printf("tolerance [0.0,1.0] = %f\n",tolerance);
		break;
	case '0': case ')': //increment size of tolerance on pixel region radius edges
		tolerance += 0.005;
		if(tolerance > 1.0)
			tolerance = 1.0;
		printf("tolerance [0.0,1.0] = %f\n",tolerance);
		break;
	case ';': case ':': //decrement brightness of picture
		luminance_boost -= 0.005;
		if(luminance_boost < -1.0)
			luminance_boost = -1.0;
		printf("luminance_boost [0.0,1.0] = %f\n",luminance_boost);
		break;
	case '\'': case '"': //increment brightness of picture
		luminance_boost += 0.005;
		if(luminance_boost > 1.0)
			luminance_boost = 1.0;
		printf("luminance_boost [0.0,1.0] = %f\n",luminance_boost);
		break;
	case ',': case '<': //decrement available shades of color for LED display
		luminance_steps--;
		if(luminance_steps < 0)
			luminance_steps = 0;
		printf("luminance_steps = %i\n",luminance_steps);
		break;
	case '.': case '>': //increment available shades of color for LED display
		luminance_steps++;
		printf("luminance_steps = %i\n",luminance_steps);
		break;
	case 'x': case 'X': //decrement color intensity
		color_boost -= 0.005;
		if(color_boost < 0.0)
			color_boost = 0.0;
		printf("color_boost [0.0,1.0] = %f\n",color_boost);
		break;
	case 'c': case 'C': //increment color intensity
		color_boost += 0.005;
		if(color_boost > 1.0)
			color_boost = 1.0;
		printf("color_boost [0.0,1.0] = %f\n",color_boost);
		break;
	case 'o': case 'O': //decrement amount of burnt out lights in LED display
		burnt_out_percent -= 0.005;
		if(burnt_out_percent < 0.0)
			burnt_out_percent = 0.0;
		printf("burnt_out_percent [0.0,1.0] = %f\n",burnt_out_percent);
		break;
	case 'p': case 'P': //increment amount of burnt out lights in LED display
		burnt_out_percent += 0.005;
		if(burnt_out_percent > 1.0)
			burnt_out_percent = 1.0;
		printf("burnt_out_percent [0.0,1.0] = %f\n",burnt_out_percent);
		break;
	case '1': //setup preset 1
		drawExample1 = !drawExample1;
		drawExample2 = false;
		if(!drawExample1)
			keyHandler('r',0,0);
		else {
			current_texture = FLAG_ID;
			current_type = BILL_IV;
			billboard_shader = billboard4_shader;
			billboard_texture_count = 2;

			//set up uniform configuration for this preset
			pixel_size = billboards[current_texture].lcf * 3;
			tolerance = 0.3;
			pixel_radius = 0.6;
			color_boost = 0.03;
			burnt_out_percent = 0.2;
			luminance_boost = -0.25;
			luminance_steps = 16;
		}
		break;
	case '2': //setup preset 2
		drawExample1 = false;
		drawExample2 = !drawExample2;
		if(!drawExample2)
			keyHandler('r',0,0);
		else {
			current_texture = FLAG_ID;
			current_type = BILL_IV;
			billboard_shader = billboard4_shader;
			billboard_texture_count = 2;

			//set up uniform configuration for this preset
			pixel_size = billboards[current_texture].lcf * 3;
			pixel_radius = 0.6;
			burnt_out_percent = 0.2;
			tolerance = 0.5;
			luminance_boost = 0.13;
			luminance_steps = 16;
			color_boost = .075;
		}
		break;
	case 'r': //reset demo
		pixel_size = billboards[current_texture].lcf * 3;
		tolerance = 0.3;
		pixel_radius = 0.6;
		luminance_steps = 12;
		luminance_boost = 0.0;
		color_boost = 0.03;
		burnt_out_percent = 0.2;
		billboard_texture_count = MAX_BILLBOARD_TEXTURES;
		current_type = BILL_I;
		billboard_shader = billboard1_shader;
		printf("Billboard I\n");
		printf("pixelSize = %i\n",pixel_size);
		printf("pixelRadius [0.0,1.0] = %f\n",pixel_radius);
		printf("tolerance [0.0,1.0] = %f\n",tolerance);		
		printf("luminance_steps = %i\n",luminance_steps);
		printf("luminance_boost [-1.0,1.0] = %f\n",luminance_boost);
		printf("color_boost [0.0,1.0] = %f\n",color_boost);
		printf("burnt_out_percent [0.0,1.0] = %f\n\n",burnt_out_percent);
		drawExample1 = false;
		drawExample2 = false;
		break;
	case 't': //cycle through textures
		current_texture++;
		if(current_texture >= billboard_texture_count)
			current_texture = 0;
		pixel_size = billboards[current_texture].lcf * 3;
		printf("pixelSize = %i\n",pixel_size);
		break;
	case 'b': //cycle through billboard shaders
		current_type++;
		if(current_type >= BILLBOARD_TYPE_COUNT)
			current_type = 0;
		switch(current_type) {
			case BILL_I:
				printf("Billboard I\n");
				printf("pixelSize = %i\n\n",pixel_size);
				billboard_shader = billboard1_shader;
				break;
			case BILL_II:
				printf("Billboard II\n");
				printf("pixelSize = %i\n",pixel_size);
				printf("pixelRadius [0.0,1.0] = %f\n",pixel_radius);
				printf("tolerance [0.0,1.0] = %f\n\n",tolerance);	
				billboard_shader = billboard2_shader;
				break;
			case BILL_III:
				printf("Billboard III\n");
				printf("pixelSize = %i\n",pixel_size);
				printf("pixelRadius [0.0,1.0] = %f\n",pixel_radius);
				printf("tolerance [0.0,1.0] = %f\n",tolerance);		
				printf("luminance_steps = %i\n",luminance_steps);
				printf("luminance_boost [-1.0,1.0] = %f\n\n",luminance_boost);
				billboard_shader = billboard3_shader;
				break;
			case BILL_IV:
				printf("Billboard IV\n");
				printf("pixelSize = %i\n",pixel_size);
				printf("pixelRadius [0.0,1.0] = %f\n",pixel_radius);
				printf("tolerance [0.0,1.0] = %f\n",tolerance);		
				printf("luminance_steps = %i\n",luminance_steps);
				printf("luminance_boost [-1.0,1.0] = %f\n",luminance_boost);
				printf("color_boost [0.0,1.0] = %f\n",color_boost);
				printf("burnt_out_percent [0.0,1.0] = %f\n\n",burnt_out_percent);
				billboard_shader = billboard4_shader;
				break;
		}
		break;
	}
	glutPostRedisplay();
}

//if you don't have a mouse you can still look around with arrow keys!
void specialKeyHandler(int key, int x, int y)
{
	switch(key)
	{
	case GLUT_KEY_UP:
		myView.pitch(3);
		break;
	case GLUT_KEY_DOWN:
		myView.pitch(-3);
		break;
	case GLUT_KEY_LEFT:
		myView.turn(3);
		break;
	case GLUT_KEY_RIGHT:
		myView.turn(-3);
		break;
	}
}

//init a camera view change with a button click
void mouseButtonHandler(int button, int state, int x, int y)
{
	if(button == GLUT_LEFT_BUTTON) {
		if(state == GLUT_DOWN) {
			xPrev = x;
			yPrev = y;
			dragging = true;
		}
		else //state == GLUT_UP
			dragging = false;
	}
	glutPostRedisplay();
}

//drag the mouse to update the camera
void mouseMotionHandler(int x, int y)
{
	int xDiff, yDiff;

	//process mouse motion and adjust camera view accordingly
	if(dragging) {
		xDiff = x - xPrev;
		yDiff = y - yPrev;
		if(xDiff != 0)
			myView.turn(-xDiff);
		if(yDiff != 0)
			myView.pitch(-yDiff);
		xPrev = x;
		yPrev = y;
		glutPostRedisplay();
	}
}

/****************************Initialization and Other functions*******************************/
//load any bitmap textures for use.  Store the handle in textures[texID]
//load any bitmap textures for use.  Store the handle in textures[texID]
bool loadTexture(int texID, char* strFileName, bool isBillboard = false)
{
	glEnable(GL_TEXTURE_2D);
	AUX_RGBImageRec *pImage = NULL;
	FILE *pFile = NULL;

	if(!strFileName) 
		return false;

	// Open a file pointer to the BMP file and check if it was found and opened 
	if((pFile = fopen(strFileName, "rb")) == NULL) 
	{
		return NULL;
	}

	// Load the bitmap using the aux function stored in glaux.lib
	pImage = auxDIBImageLoad(strFileName);				

	// Make sure valid image data was given to pImage, otherwise return false
	if(pImage == NULL)								
		return false;

	if(isBillboard) {
		billboards[texID].x = pImage->sizeX;
		billboards[texID].y = pImage->sizeY;
	}

	// Generate a texture with the associative texture ID stored in the array
	glGenTextures(1, &textures[texID]);

	// This sets the alignment requirements for the start of each pixel row in memory.
	glPixelStorei (GL_UNPACK_ALIGNMENT, 1);

	// Bind the texture to the texture arrays index and init the texture
	glBindTexture(GL_TEXTURE_2D, textures[texID]);
	
	// Build Mipmaps (builds different versions of the picture for distances - looks better)
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, pImage->sizeX, 
					  pImage->sizeY, GL_RGB, GL_UNSIGNED_BYTE, pImage->data);

	//Assign the mip map levels and texture info
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// Now we need to free the image data that we loaded since openGL stored it as a texture

	if (pImage)										// If we loaded the image
	{
		if (pImage->data)							// If there is texture data
		{
			free(pImage->data);						// Free the texture data, we don't need it anymore
		}

		free(pImage);								// Free the image structure
	}

	//glDisable(GL_TEXTURE_2D);
	// Return a success
	return true;
}

//returns least common factor of a & b, used to get a pixelSize step factor for the current textures resolution
int getLCF (int a, int b, int lcf = 1) {
	lcf++;
	while(lcf < a && lcf < b) {
		if((a % lcf) == 0 && (b % lcf) == 0)
			return lcf;
		else
			return getLCF(a, b, lcf);
	}
}

//initialize other data
void InitData()
{	
	//initialize camera view
	myView.positionCamera(0.0,20.0,60.0, 0.0,15.0,0.0, 0.0,1.0,0.0);

	//load billboard model used in presets
	billboardObjFile = obj_add_file("./Models/billboard.obj");

	//load our bmps
	loadTexture(FLAG_ID,".\\Textures\\flagcross.bmp", true);
	loadTexture(WWBD_ID,".\\Textures\\wwbd.bmp", true);
	loadTexture(HOMESTAR_ID,".\\Textures\\homestar.bmp", true);
	loadTexture(LOGO_ID,".\\Textures\\logoGray.bmp", true);
	loadTexture(NOAH_ID,".\\Textures\\noah.bmp", true);
	loadTexture(NOISE_ID,".\\Textures\\noise.bmp");
	loadTexture(DAY_GRASS_ID,".\\Textures\\dayGrass.bmp");
	loadTexture(DAY_ID,".\\Textures\\dayBackground.bmp");
	loadTexture(NIGHT_GRASS_ID,".\\Textures\\nightGrass.bmp");
	loadTexture(NIGHT_ID,".\\Textures\\nightBackground.bmp");
	loadTexture(METAL_ID,".\\Textures\\metal.bmp");
	loadTexture(NEON_ON_ID,".\\Textures\\neonOn.bmp");
	loadTexture(NEON_INDEX,".\\Textures\\neonOff.bmp",true);

	//set neon texture to off initially
	textures[NEON_OFF_ID] = textures[NEON_INDEX];

	//assign lcf for billboard textures
	billboards[FLAG_ID].lcf = getLCF(billboards[FLAG_ID].x,billboards[FLAG_ID].y);
	billboards[WWBD_ID].lcf = getLCF(billboards[WWBD_ID].x,billboards[WWBD_ID].y);
	billboards[HOMESTAR_ID].lcf = getLCF(billboards[HOMESTAR_ID].x,billboards[HOMESTAR_ID].y);
	billboards[LOGO_ID].lcf = getLCF(billboards[LOGO_ID].x,billboards[LOGO_ID].y);
	billboards[NOAH_ID].lcf = getLCF(billboards[NOAH_ID].x,billboards[NOAH_ID].y);
	billboards[NEON_INDEX].lcf = getLCF(billboards[NEON_INDEX].x,billboards[NEON_INDEX].y);
	
	//set some initial operation modes for the demo
	billboard_shader = billboard1_shader;
	current_type = BILL_I;
	current_texture = FLAG_ID;
	pixel_size = billboards[current_texture].lcf * 3;

	printf("Billboard I\n");
	printf("pixelSize = %i\n\n",pixel_size);
}

//set up glut
void InitGlut (int a, char **b) {
	glutInit(&a, b);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowPosition(100,100);
	glutInitWindowSize(700,700);
	glutCreateWindow("Billboard Tutorial");

	glutDisplayFunc(display);
	glutReshapeFunc(changeSize);
	glutSpecialFunc(specialKeyHandler);
	glutKeyboardFunc(keyHandler);
	glutMouseFunc(mouseButtonHandler);
	glutMotionFunc(mouseMotionHandler);
}

//setup gl
void InitGL () {
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glClearColor(0.0,0.0,0.0,1.0);

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);

	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f,glutGet(GLUT_WINDOW_WIDTH)/glutGet(GLUT_WINDOW_HEIGHT), 0.9f ,100.0f);

	glDepthFunc (GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glDisable(GL_LIGHTING);

	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix
}

//setup all of our billboard shaders
void InitShaders()
{
	/****************************************************
						billboard1 Shader
	****************************************************/
	char* billboard1_vert = NULL;
	char* billboard1_frag = NULL;
	billboard1_vert = readShaderFile("./Shaders/billboard1.vert");
	billboard1_frag = readShaderFile("./Shaders/billboard1.frag");

	// create two shader objects and return their handles
    v = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
    f = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);    

    const char * const_billboard1_vert = billboard1_vert;
    const char * const_billboard1_frag = billboard1_frag;

    // sets the source code of the shaders to be the source code of the strings
    glShaderSourceARB(v, 1, &const_billboard1_vert, NULL);
    glShaderSourceARB(f, 1, &const_billboard1_frag, NULL);

    delete billboard1_vert, billboard1_frag;

    // compile the shaders - this can succeed or fail - more on this later
    glCompileShaderARB(v);
    glCompileShaderARB(f);

    // create a program object to link the shaders together and return its handle
    billboard1_shader = glCreateProgramObjectARB();

    // attach shader objects to the program object
    glAttachObjectARB(billboard1_shader,v);
    glAttachObjectARB(billboard1_shader,f);

    // link the program object - this can succeed or fail - more on this later
    glLinkProgramARB(billboard1_shader);

	/****************************************************
						billboard2 Shader
	****************************************************/
	char* billboard2_vert = NULL;
	char* billboard2_frag = NULL;
	billboard2_vert = readShaderFile("./Shaders/billboard2.vert");
	billboard2_frag = readShaderFile("./Shaders/billboard2.frag");

	// create two shader objects and return their handles
    v = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
    f = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);    

    const char * const_billboard2_vert = billboard2_vert;
    const char * const_billboard2_frag = billboard2_frag;

    // sets the source code of the shaders to be the source code of the strings
    glShaderSourceARB(v, 1, &const_billboard2_vert, NULL);
    glShaderSourceARB(f, 1, &const_billboard2_frag, NULL);

    delete billboard2_vert, billboard2_frag;

    // compile the shaders - this can succeed or fail - more on this later
    glCompileShaderARB(v);
    glCompileShaderARB(f);

    // create a program object to link the shaders together and return its handle
    billboard2_shader = glCreateProgramObjectARB();

    // attach shader objects to the program object
    glAttachObjectARB(billboard2_shader,v);
    glAttachObjectARB(billboard2_shader,f);

    // link the program object - this can succeed or fail - more on this later
    glLinkProgramARB(billboard2_shader);

	/****************************************************
						billboard3 Shader
	****************************************************/
	char* billboard3_vert = NULL;
	char* billboard3_frag = NULL;
	billboard3_vert = readShaderFile("./Shaders/billboard3.vert");
	billboard3_frag = readShaderFile("./Shaders/billboard3.frag");

	// create two shader objects and return their handles
    v = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
    f = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);    

    const char * const_billboard3_vert = billboard3_vert;
    const char * const_billboard3_frag = billboard3_frag;

    // sets the source code of the shaders to be the source code of the strings
    glShaderSourceARB(v, 1, &const_billboard3_vert, NULL);
    glShaderSourceARB(f, 1, &const_billboard3_frag, NULL);

    delete billboard3_vert, billboard3_frag;

    // compile the shaders - this can succeed or fail - more on this later
    glCompileShaderARB(v);
    glCompileShaderARB(f);

    // create a program object to link the shaders together and return its handle
    billboard3_shader = glCreateProgramObjectARB();

    // attach shader objects to the program object
    glAttachObjectARB(billboard3_shader,v);
    glAttachObjectARB(billboard3_shader,f);

    // link the program object - this can succeed or fail - more on this later
    glLinkProgramARB(billboard3_shader);

	/****************************************************
						billboard4 Shader
	****************************************************/
	char* billboard4_vert = NULL;
	char* billboard4_frag = NULL;
	billboard4_vert = readShaderFile("./Shaders/billboard4.vert");
	billboard4_frag = readShaderFile("./Shaders/billboard4.frag");

	// create two shader objects and return their handles
    v = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
    f = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);    

    const char * const_billboard4_vert = billboard4_vert;
    const char * const_billboard4_frag = billboard4_frag;

    // sets the source code of the shaders to be the source code of the strings
    glShaderSourceARB(v, 1, &const_billboard4_vert, NULL);
    glShaderSourceARB(f, 1, &const_billboard4_frag, NULL);

    delete billboard4_vert, billboard4_frag;

    // compile the shaders - this can succeed or fail - more on this later
    glCompileShaderARB(v);
    glCompileShaderARB(f);

    // create a program object to link the shaders together and return its handle
    billboard4_shader = glCreateProgramObjectARB();

    // attach shader objects to the program object
    glAttachObjectARB(billboard4_shader,v);
    glAttachObjectARB(billboard4_shader,f);

    // link the program object - this can succeed or fail - more on this later
    glLinkProgramARB(billboard4_shader);
}

//entry point for our program
int main(int argc, char* argv[])
{
	printf("**********************************************************\n");
	printf("LED Billboard Tutorial Demo by: Jason Gorski\n");
	printf("Please refer to included ReadMe.txt file for commands!\n");
	printf("**********************************************************\n");
	printf("Loading Bitmaps...\n\n");

	InitGlut(argc, argv);
	InitGL();

	glewInit();
	if(!GLEW_ARB_vertex_program || !GLEW_ARB_fragment_program)
	{
		exit(1);
	}

	InitShaders();
	InitData();

	glutMainLoop();
	return 0;
}

